use std::{collections::HashMap, path::PathBuf};

use chrono::{DateTime, Utc};
use color_eyre::eyre::{Report, Result};
use itertools::Itertools;
use serde::Deserialize;
use walkdir::WalkDir;

use crate::storage;

const PROJECT_ID: &str = "JR0bkFKa";

#[derive(Debug, Deserialize)]
pub struct ProjectVersion {
    pub name: String,
    pub version_number: String,
    game_versions: Vec<String>,
    files: Vec<VersionFile>,
    date_published: DateTime<Utc>,
}

#[derive(Debug, Deserialize)]
pub struct VersionFile {
    url: String,
    filename: String,
    hashes: VersionFileHashes,
}

#[derive(Debug, Deserialize)]
pub struct VersionFileHashes {
    sha512: String,
}

#[derive(Debug, Deserialize)]
#[serde(rename_all = "camelCase")]
struct ModpackIndex {
    files: Vec<ModpackFile>,
    dependencies: HashMap<String, String>,
}

#[derive(Debug, Deserialize)]
struct ModpackFile {
    path: PathBuf,
    hashes: VersionFileHashes,
    downloads: Vec<String>,
}

impl ProjectVersion {
    pub async fn download_file(&self) -> Result<PathBuf> {
        let version_file = &self.files[0];

        storage::download_to_cache(
            &version_file.url,
            &version_file.filename,
            &version_file.hashes.sha512,
        )
        .await
    }

    pub async fn install_modpack(&self) -> Result<PathBuf> {
        let mrpack_path = self.download_file().await?;
        let decompressed = storage::decompress_zip(&mrpack_path)?;

        let index_path = decompressed.join("modrinth.index.json");
        let index = serde_json::from_slice::<ModpackIndex>(&std::fs::read(&index_path)?)?;

        let version_dir = storage::get_version_directory(&self.game_versions[0])?;

        for file in index.files {
            let Some(filename) = file.path.file_name() else {
                tracing::warn!("path {} does not have a filename", file.path.display());
                continue;
            };

            let download_path = storage::download_to_cache(
                &file.downloads[0],
                &filename.to_string_lossy(),
                &file.hashes.sha512,
            )
            .await?;

            let final_path = version_dir.join(file.path);
            storage::move_file(download_path, &final_path)?;

            tracing::info!("downloaded file {}", final_path.display());
        }

        let overrides_path = decompressed.join("overrides");
        for entry in WalkDir::new(&overrides_path) {
            let entry = entry?;

            if entry.path().is_file() {
                let stripped = entry.path().strip_prefix(&overrides_path)?;
                let final_path = version_dir.join(stripped);
                storage::move_file(entry.path(), &final_path)?;

                tracing::info!("added override {}", final_path.display());
            }
        }

        Ok(version_dir)
    }
}

pub async fn get_versions() -> Result<Vec<ProjectVersion>> {
    let url = format!("https://api.modrinth.com/v2/project/{PROJECT_ID}/version");

    crate::HTTP_CLIENT
        .get(url)
        .send()
        .await?
        .json()
        .await
        .map_err(Report::from)
}

pub trait VersionsExt {
    fn available_minecraft_versions(&self) -> Vec<&String>;
    fn latest_for_mc_version(&self, mc_version: &str) -> Option<&ProjectVersion>;
}

impl VersionsExt for Vec<ProjectVersion> {
    fn available_minecraft_versions(&self) -> Vec<&String> {
        // we can safely assume that there is only one game_version per project version since it's a modpack
        self.iter().map(|v| &v.game_versions[0]).unique().collect()
    }

    fn latest_for_mc_version(&self, mc_version: &str) -> Option<&ProjectVersion> {
        self.iter()
            .filter(|v| v.game_versions.contains(&mc_version.to_owned()))
            .sorted_by_key(|v| v.date_published)
            .next_back()
    }
}
