use std::{
    fs::File,
    path::{Path, PathBuf},
    sync::LazyLock,
};

use color_eyre::eyre::{bail, eyre, OptionExt, Report, Result};
use directories::ProjectDirs;
use sha2::{Digest, Sha512};
use zip::ZipArchive;

static DIRS: LazyLock<ProjectDirs> = LazyLock::new(|| {
    ProjectDirs::from("net", "uku3lig", "uklient")
        .expect("could not instantiate project directories!")
});

pub fn make_directories() -> std::io::Result<()> {
    std::fs::create_dir_all(DIRS.data_dir())?;
    std::fs::create_dir_all(DIRS.cache_dir())?;

    Ok(())
}

pub async fn download_to_cache(url: &str, filename: &str, sha512: &str) -> Result<PathBuf> {
    let path = DIRS.cache_dir().join(filename);

    let response = crate::HTTP_CLIENT.get(url).send().await?;
    let mut bytes = &*response.bytes().await?; // deref to get &[u8]

    let expected_hash = hex::decode(sha512)?;
    let computed_hash = {
        let mut sha512 = Sha512::new();
        sha512.update(bytes);
        sha512.finalize()
    };

    if expected_hash[..] != computed_hash[..] {
        bail!(
            "hash mismatch! expected: {}, got {}",
            hex::encode(expected_hash),
            hex::encode(computed_hash),
        )
    }

    let mut file = File::create(&path)?;
    std::io::copy(&mut bytes, &mut file)?;

    Ok(path)
}

pub fn decompress_zip(path: &Path) -> Result<PathBuf> {
    let dir = path.parent().ok_or_eyre("could not find parent")?;
    let name = path.file_stem().ok_or_eyre("filename does not exist")?;
    let extract_dir = dir.join(name);

    let file = File::open(path)?;
    let mut archive = ZipArchive::new(file)?;
    archive.extract(&extract_dir)?;

    Ok(extract_dir)
}

pub fn get_version_directory(version: &str) -> Result<PathBuf> {
    let path = DIRS.data_dir().join("versions").join(version);
    std::fs::create_dir_all(&path)?;
    Ok(path)
}

pub fn move_file<S: AsRef<Path>, D: AsRef<Path>>(source: S, dest: D) -> Result<()> {
    let parent = dest
        .as_ref()
        .parent()
        .ok_or_else(|| eyre!("path {} does not have a parent", dest.as_ref().display()))?;

    std::fs::create_dir_all(parent)?;
    std::fs::rename(source, dest).map_err(Report::from)
}
